// Rudimentary serial port drivers for Metrowerks C

// Copyright  1997 by Francis J. Deck.
// It is OK to use this code, but only "at your own risk."

// This is a modest but fairly rigorous serial port driver. I wrote it to
// use in some software that I publish as Freeware with all of the usual
// disclaimers. An unusual feature of this driver is that it uses a single
// long-integer reference number for the ports, which your program must
// store and maintain. The reason for this is to allow incorporation of
// this code into code resources, which are not allowed to have global
// variables.

// All routines have error checking. Only the routine serialOpen() actually
// closes the ports if it encounters an error. Your program _must_ make sure
// that the ports are properly closed before terminating, or the open
// serial handlers will simply lurk on your Mac until you restart.

// We use only the default 64-byte buffers. That's OK for my purposes,
// since my data aquisition boards typically process commands with no more
// than half a dozen characters.

// There is no error checking for valid parameters. Also, the only routine
// that closes the ports upon encountering an error is serialOpen(). In
// general, your program must close the ports before terminating, or risk
// leaving an open port hanging on the system until restart.

#include <serial.h>
#include "serialDrivers.h"

// data structure used to combine input and output reference numbers

union dualdata
	{
	long lg; // combined number
	struct {short inrefnum, outrefnum;} sg; // separate values
	};
		
// Close the serial port

void serialClose
	(long refnum) // combined input/output reference number
	
	{
	union dualdata dd; // driver reference numbers
	
	dd.lg = refnum; // separate the reference numbers
	CloseDriver (dd.sg.inrefnum); // close the input driver
	CloseDriver (dd.sg.outrefnum); // close the output driver
	}
	
// Open a serial port

OSErr serialOpen
	(long *refnum, // reference number to be assigned a value here
	long nport, // which port, 0 = Modem, 1 = Printer
	long baud) // standard baud rate in Hz
	
	{
	
	union dualdata dd; // driver reference numbers
	short baudval; // baud rate in "mac units"
	SerShk ss; // handshake parameters
		
	switch (nport)  // interpret user's port selection and open driver
		{
		case 0: // modem port
			
		if (OpenDriver("\p.AOut", &dd.sg.outrefnum) != noErr) return yesErr;
			
		if (OpenDriver("\p.Ain", &dd.sg.inrefnum) != noErr) 
			{
			CloseDriver (dd.sg.outrefnum); // close hanging output driver
			return yesErr;
			}
		break;
		
		case 1: // printer port
			
		if (OpenDriver("\p.Bout", &dd.sg.outrefnum) != noErr) return yesErr;
			
		if (OpenDriver("\p.Bin", &dd.sg.inrefnum) != noErr) 
			{
			CloseDriver (dd.sg.outrefnum); // close hanging output driver
			return yesErr;
			}
			
		break;
		}
		
	switch (baud) // interpret user's baud rate selection 
		{
		case 300: { baudval = baud300; break; }
		case 600: { baudval = baud600; break; }
		case 1200: { baudval = baud1200; break; }
		case 1800: { baudval = baud1800; break; }
		case 2400: { baudval = baud2400; break; }
		case 3600: { baudval = baud3600; break; }
		case 4800: { baudval = baud4800; break; }
		case 7200: { baudval = baud7200; break; }
		case 9600: { baudval = baud9600; break; }
		case 14400: { baudval = 6; break; }
		case 19200: { baudval = baud19200; break; }
		case 28800: { baudval = 2; break; }
		case 38400: { baudval = baud38400; break; }
		case 57600: { baudval = baud57600; break; }		
		}
	
	// reset the ports
	
	if (SerReset(dd.sg.inrefnum, baudval + stop10 + noParity + data8) != noErr)
		{
		serialClose(dd.lg);
		return yesErr;
		}
		
	if (SerReset(dd.sg.outrefnum, baudval + stop10 + noParity + data8) != noErr)
		{
		serialClose(dd.lg);
		return yesErr;
		}
		
	// set up handshaking parameters
	
	ss.fXOn = 0; // don't use XON/XOFF handshaking
	ss.xOn = 19; // usual XON character (which is irrelevant here)
	ss.xOff = 17; // usual XOFF character (which is irrelevant here)
	ss.fCTS = 0; // don't use CTS (input) handshaking either
	ss.errs = 0; // ignore input errors
	ss.evts = 0; // there are no device driver events
	ss.fDTR = 0; // don't use DTR (output) handshaking either
		
	if (SerHShake(dd.sg.inrefnum, &ss) != noErr)
		{
		serialClose(dd.lg);
		return yesErr;
		}
	
	if (SerHShake(dd.sg.outrefnum, &ss) != noErr)
		{
		serialClose(dd.lg);
		return yesErr;
		}
		
	*refnum = dd.lg;
	
	return noErr;
	}
	
// Get the number of characters available in the receive buffer

OSErr serialNAvail
	(
	long refnum, // reference number assigned by serialOpen()
	long *navail // numbre of characters available
	)
	
	{
	union dualdata dd; // driver reference numbers
	
	dd.lg = refnum;
	
	if (SerGetBuf(dd.sg.inrefnum, navail) != noErr)
		{
		serialClose(refnum);
		return yesErr;
		}
	return noErr;
	}
	
// Get the next available character from the serial port

OSErr serialReadChar
	(long refnum, // reference number
	unsigned char *c) // incoming character
	
	{
	union dualdata dd; // driver reference numbers
	long cn; // byte count
	
	dd.lg = refnum;
	
	// defensively, check that characters are available
	
	if (SerGetBuf(dd.sg.inrefnum, &cn) != noErr) return yesErr;
	if (cn < 1)
		{
		*c = 0;
		return noErr;
		}
	
	// now read the character
	
	cn = 1; // we want to read exactly one character	
	if (FSRead(dd.sg.inrefnum, &cn, c) != noErr) return yesErr;	
	return noErr;
	}
	
// Send a character to the serial port

OSErr serialSendChar
	(long refnum, // reference number
	unsigned char c) // character to be sent
	
	{
	union dualdata dd; // port reference numbers
	long cn; // byte count
	
	dd.lg = refnum;
	cn = 1; // we want to write exactly one byte
	if (FSWrite(dd.sg.outrefnum, &cn, &c) != noErr) return yesErr;
	return noErr;
	}
	
// Flush the receive buffer

OSErr serialFlush
	(long refnum) // reference number
	
	{
	union dualdata dd; // port reference number
	long cn = 0; // byte count
	char c; // dummy character being read
	
	dd.lg = refnum;
	
	// get number of characters available
	
	if (SerGetBuf(dd.sg.inrefnum, &cn) != noErr) return yesErr;
	
	// loop 'til buffer is empty
	
	while (cn > 0)
		{
		cn = 1; // one character at a time
		
		// read a character
		
		if (FSRead(dd.sg.inrefnum, &cn, &c) != noErr) return yesErr;
		
		// re-measure number of characters available
		
		if (SerGetBuf(dd.sg.inrefnum, &cn) != noErr) return yesErr;
		}
	return noErr;
	}
	
// Timeout loop waiting for fixed number of characters

OSErr serialTimeout
	(long refnum, // reference number
	long howmany, // how many bytes to wait for
	long nticks) // how many 1/60 second intervals to wait for
	
	{
	long timeout; // the witchin' hour
	long nchars = 0; // actual number of characters available
	
	timeout = TickCount() + nticks;
		
	while (nchars < howmany)
		{
		if (TickCount() > timeout) return yesErr;
		if (serialNAvail(refnum, &nchars) != noErr) return yesErr;
		}
		
	return noErr;
	}
	
// Wait for data transmission to finish

OSErr serialXmitWait
	(long refnum)
	
	{
	union dualdata dd;
	SerStaRec ss;
	
	dd.lg = refnum;
	if (SerStatus (dd.sg.outrefnum, &ss) != noErr) return yesErr
	
	while (ss.wrPend != 0)
		if (SerStatus (dd.sg.outrefnum, &ss) != noErr) return yesErr;
		
	return noErr;
	}