/***********************************************************************
 *  avrp - Atmel AVR programming software to use with Atmel's
 *         serial-port programmers.
 *  Portions are Copyright (C) 1997-1998 Jon Anders Haugum
 *  Copyright (C) 1999 Ben Hitchcock
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; see the file COPYING.  If not, write to
 *  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 *  Boston, MA 02111-1307, USA.
 *
 *
 *  Author of avrp can be reached at:
 *     email: jonah@colargol.tihlde.hist.no
 *     www: http://www.colargol.tihlde.hist.no/~jonah/el/avrp.html
 *     Postal address: Jon Anders Haugum
 *                     vre Mllenbergsgt 52
 *                     7014 Trondheim
 *  Author of MacAVRpa can be reached at:
 *     email: ben@wollongong.apana.org.au
 *     Postal address: Ben Hitchcock
 *     C/0  Don Hitchcock
 *          "Fangorn", Old Inverell Rd
 *          Armidale
 *          NSW 2350
 *          AUSTRALIA
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "avrp.h"
#include "avra.h"


const char *Title = "avrp 1.0 beta 3 (28 September 1998)\n"
                    "Copyright (C) 1997-1998 Jon Anders Haugum\n"
                    "   avrp comes with NO WARRANTY, to the extent permitted by law.\n"
                    "   You may redistribute copies of avrp under the terms\n"
                    "   of the GNU General Public License.\n"
                    "   For more information about these matters, see the files named COPYING.\n";

const char *Usage =
#ifdef __amigaos__
	"usage: avrp [-s <serial device:serial unit>] [-a <chip type>] [-prv]\n"
#else
	"usage: avrp [-s <serial port>] [-a <chip type>] [-prv]\n"
#endif
	"            [-f <filename>] [-e <filename>] [-l <lockmode>]\n"
        "            [--enable <fuse(s)>] [--disable <fuse(s)>]\n"
	"            [-d <deffile>] [-qIh] [--version] [--help]\n"
	"\n"
#ifdef __amigaos__
	"   -s : Specifies which serial port and unit to use (e.g. serial.device:0)\n"
#elif _WIN32
	"   -s : Specifies which serial port to use (e.g. COM2)\n"
#else
	"   -s : Specifies which serial port to use (e.g. /dev/ttyS1)\n"
#endif
	"   -a : Specifies chip to program/read (use '-a list' to get a list)\n"
	"   -p : Program a chip (-f or -e must be specified)\n"
	"   -r : Read a chip    (-f or -e must be specified)\n"
	"   -v : Verify         (-f or -e must be specified)\n"
	"   -f : Flashfile to write/read\n"
	"   -e : Eepromfile to write/read\n"
	"   -l : Lockmode (use '-l list' to get a list of modes)\n"
	"   -q : Quiet. Do not print out progress (little speedup)\n"
	"   -d : Specifies name and path to avrp.def\n"
	"   -I : Ignore chip signature\n"
	"   --enable   : Enable fuse(s)\n"
	"   --disable  : Disable fuse(s)\n"
	"   --autoinc  : Programmer support automatic address incrementation\n"
	"   --version  : Version information\n"
	"   --help, -h : This help text\n"
	"\n"
	"Report bugs to jonah@colargol.tihlde.hist.no\n";


int avrpmain(int argc, char *argv[])
	{
	
	int ShowUsage = False;
	char *temp_env = NULL;
	struct avrProgramInfo *PI;
	struct avrpargs *theargs;
	char* defFile;
	
	boolean * theBool, * tempBool;

	theBool = new boolean;
	defFile = new char[40];
	
	strcpy(defFile, DEFFILENAME);
	
	*theBool = (boolean) 0;

	theargs = alloc_avrpargs(P_ARG_COUNT);
	if(theargs)
		{
		define_avrparg(theargs, ARG_SERIALPORT, ARGTYPE_STRING,       's', NULL,      getenv("AVRP_SERIALPORT"));
		define_avrparg(theargs, ARG_AVRTYPE   , ARGTYPE_STRING,       'a', NULL,      getenv("AVRP_AVRTYPE"));
		define_avrparg(theargs, ARG_PROG      , ARGTYPE_BOOLEAN,      'p', NULL,      0);
		define_avrparg(theargs, ARG_READ      , ARGTYPE_BOOLEAN,      'r', NULL,     0);
		define_avrparg(theargs, ARG_VERIFY    , ARGTYPE_BOOLEAN,      'v', NULL,      0);
		define_avrparg(theargs, ARG_FLASH     , ARGTYPE_STRING,       'f', NULL,      NULL);
		define_avrparg(theargs, ARG_EEPROM    , ARGTYPE_STRING,       'e', NULL,      NULL);
		define_avrparg(theargs, ARG_QUICK     , ARGTYPE_BOOLEAN,      'q', NULL,     0);
// Commented out cos we're on a mac.  Environment variables?  What are they?
		temp_env = getenv("AVRP_DEFFILE");
//		define_avrparg(theargs, ARG_DEFFILE   , ARGTYPE_STRING,       'd', NULL,      temp_env ? temp_env : DEFFILENAME);
		define_avrparg(theargs, ARG_DEFFILE   , ARGTYPE_STRING,       'd', NULL,      defFile);
		define_avrparg(theargs, ARG_IGNORESIG , ARGTYPE_BOOLEAN,      'I', NULL,      0);
		define_avrparg(theargs, ARG_LOCKMODE  , ARGTYPE_STRING,       'l', NULL,      NULL);
		define_avrparg(theargs, ARG_ENABLE    , ARGTYPE_STRING_MULTI,   0, "enable",  NULL);
		define_avrparg(theargs, ARG_DISABLE   , ARGTYPE_STRING_MULTI,   0, "disable", NULL);
		define_avrparg(theargs, ARG_AUTOINC   , ARGTYPE_BOOLEAN,        0, "autoinc", 0);
		define_avrparg(theargs, P_ARG_VER       , ARGTYPE_BOOLEAN,        0, "version", 0);
		define_avrparg(theargs, P_ARG_HELP      , ARGTYPE_BOOLEAN,      'h', "help",    0);
		if(read_avrpargs(theargs, argc, argv))
			{
			tempBool = (boolean *)GET_avrparg(theargs, P_ARG_HELP);
			if(!GET_avrparg(theargs, P_ARG_HELP) && ((argc != 1) || GET_avrparg(theargs, ARG_SERIALPORT)))
				{
				tempBool = (boolean *)GET_avrparg(theargs, P_ARG_VER);

				if(!GET_avrparg(theargs, P_ARG_VER))
					{
					PI = GetavrpPI(theargs);
					if(PI)
						{
						if(GET_avrparg(theargs, ARG_SERIALPORT))
							{
							PI->Ser = OpenSer((char *)GET_avrparg(theargs, ARG_SERIALPORT));
							if(PI->Ser)
								{
								if(InquireHW(PI))
									{
									if(DetectAVRType(PI))
										DoTheTalk(PI);
									}
								else
								{
									doCautionAlert("\pNo Programmer found.");

//									printf("%s: No programmer found\n", (char *)GET_ARG(theargs, ARG_SERIALPORT));
								}
								CloseSer(PI->Ser);
								}
							}
						else
							{
							doCautionAlert("\pMissing argument: No serial port specified");

//							printf("Missing argument: No serial port specified\n"
#ifdef __amigaos__
//									 "Use: -s <serial device:serial unit>\n"
#else
//									 "Use: -s <serial port>\n"
#endif
//									 "Or set the environment variable: AVRP_SERIALPORT\n");
							}
						FreePI(PI);
						}
					}
				else
					{
//					printf("%s", Title);
					}
				}
			else
				ShowUsage = True;
			}
		free_avrpargs(theargs);
		}
	else
		{
		ShowUsage = True;
//		printf("\n");
		}
	if(ShowUsage)
		{
//		if(argc == 1)
//			printf("%s\n", Title);
//		printf("%s", Usage);
		}
//	CloseSer(PI->Ser);
/*	exit(EXIT_SUCCESS);*/
	return 0;
	}


int DetectAVRType(struct avrProgramInfo *PI)
	{
	int several_match = False, chiplist = False;
	char *thisChipName;
	struct ChipNode *CN;

	if(GET_avrparg(PI->avrpargs, ARG_AVRTYPE))
		{
		if(nocase_strcmp((char *)GET_avrparg(PI->avrpargs, ARG_AVRTYPE), "list"))
			{
			for(CN = PI->FirstChipNode; CN; CN = CN->Next)
				{
				thisChipName = (char *)GET_avrparg(PI->avrpargs, ARG_AVRTYPE);
				if(nocase_strstr(CN->Name, (char *)GET_avrparg(PI->avrpargs, ARG_AVRTYPE)))
					{
					if(PI->SelectedChipNode)
						several_match = True;
					PI->SelectedChipNode = CN;
//					printf("Selected chip: %s %s %s with %dKB flash\n", CN->VendorNode->Name,
//					       CN->FamilyNode->Name, CN->Name, CN->FamilyNode->FlashSize);
					}
				}
			}
		else
			chiplist = True;
		if(several_match)
			doCautionAlert("\pError: There were several chips that matched your specification." );

//			printf("Error: There were several chips that matched: %s\n"
//			       "       You must be more specific\n", (char *)GET_ARG(PI->args, ARG_AVRTYPE));
		else if(PI->SelectedChipNode)
			{
			if(PI->SelectedChipNode->Supported)
				return(True);
			else
				doCautionAlert("\pError: that chip is not supported by your programmer\n");

//				printf("Error: %s is not supported by your programmer\n", PI->SelectedChipNode->Name);
			}
		else
			{
			if(!chiplist)
			{
				char errorString[200];
				strcpy(errorString, "Error: Unable to find a match to ");
				strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_AVRTYPE));
				strcat(errorString, " in ");				
				strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_DEFFILE));
				strcat(errorString, ".");				
				c2pstr(errorString);
				doCautionAlert((unsigned char*)errorString);

//				printf("Error: Unable to find a match to '%s' in %s\n", ,
//				       (char *)GET_ARG(PI->args, ARG_DEFFILE));
			}
/*			printf("The following chips are defined:\n");
			for(CN = PI->FirstChipNode; CN; CN = CN->Next)
				{
				if(CN->Supported)
					printf("   %s (%s %s with %d kB flash)\n", CN->Name, CN->VendorNode->Name,
					       CN->FamilyNode->Name, CN->FamilyNode->FlashSize);
				else
					printf("   %s (%s %s with %d kB flash) Not supported by programmer\n",
					       CN->Name, CN->VendorNode->Name, CN->FamilyNode->Name,
					       CN->FamilyNode->FlashSize);
				}
*/
			}

		}
	else
		{

		doCautionAlert("\pMissing argument: Type of AVR must be specified.");

/*		printf("Missing argument: Type of AVR must be specified\n"
		       "Use: -a <AVR type>\n"
		       "Or set the environment variable: AVRP_AVRTYPE\n"
		       "Shortest unique abbreviation can be used\n"
		       "List of all supported chips can be obtained with: -a list\n");
*/
		}
	return(False);
	}


struct avrProgramInfo *GetavrpPI(struct avrpargs *avrpargs)
	{
	int ok = False;
	struct avrProgramInfo *PI;
	struct ChipNode *CN;
	struct VendorNode *VN;
	struct FamilyNode *FamN;
	struct ArchNode *AN;
	struct FuseNode *FuseN;
	struct LockNode *LN;

	PI = (struct avrProgramInfo *)calloc(1, sizeof(struct avrProgramInfo));
	if(PI)
		{
		PI->avrpargs = avrpargs;
		if(ReadDefFile(PI))
			{
			ok = True;
			for(FamN = PI->FirstFamilyNode; FamN; FamN = FamN->Next)
				{
				for(AN = PI->FirstArchNode; AN; AN = AN->Next)
					if(!strcmp(FamN->Name, AN->Name))
						{
						FamN->ArchNode = AN;
						break;
						}
				if(!FamN->ArchNode)
					{
					ok = False;
					char errorString[200];
					strcpy(errorString, "Error: ");
					strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_DEFFILE));
					strcat(errorString, ": Can't find architecture ");
					strcat(errorString, FamN->Name);
					c2pstr(errorString);
					doCautionAlert((unsigned char*)errorString);

//					printf("Error: %s: Can't find architecture %s\n",
//					       (char *)GET_ARG(PI->args, ARG_DEFFILE), FamN->Name);
					}
				}
			for(CN = PI->FirstChipNode; CN; CN = CN->Next)
				{
				for(VN = PI->FirstVendorNode; VN; VN = VN->Next)
					if(VN->Code == ((CN->Sig >> 16) & 0xFF))
						{
						CN->VendorNode = VN;
						break;
						}
				for(FamN = PI->FirstFamilyNode; FamN; FamN = FamN->Next)
					if(FamN->Code == ((CN->Sig >> 8) & 0xFF))
						{
						CN->FamilyNode = FamN;
						break;
						}
				if(!(CN->VendorNode && CN->FamilyNode))
					{
					ok = False;
					char errorString[200];
					strcpy(errorString, "Error: ");
					strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_DEFFILE));
					strcat(errorString, ": No vendor and family matched the signature of ");
					strcat(errorString, CN->Name);
					c2pstr(errorString);
					doCautionAlert((unsigned char*)errorString);
//					printf("Error: %s: No vendor and family matched the signature of %s\n",
//					       (char *)GET_ARG(PI->args, ARG_DEFFILE), CN->Name);
					}
				else if(!CN->VendorNode)
					{
					ok = False;
					char errorString[200];
					strcpy(errorString, "Error: ");
					strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_DEFFILE));
					strcat(errorString, ": No vendor matched the signature of ");
					strcat(errorString, CN->Name);
					c2pstr(errorString);
					doCautionAlert((unsigned char*)errorString);
//					printf("Error: %s: No vendor matched the signature of %s\n",
//					       (char *)GET_ARG(PI->args, ARG_DEFFILE), CN->Name);
					}
				else if(!CN->FamilyNode)
					{
					ok = False;
					char errorString[200];
					strcpy(errorString, "Error: ");
					strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_DEFFILE));
					strcat(errorString, ": No family matched the signature of ");
					strcat(errorString, CN->Name);
					c2pstr(errorString);
					doCautionAlert((unsigned char*)errorString);
//					printf("Error: %s: No family matched the signature of %s\n",
//					       (char *)GET_ARG(PI->args, ARG_DEFFILE), CN->Name);
					}
				if(CN->FuseType != 0)
					{
					for(FuseN = PI->FirstFuseNode; FuseN; FuseN = FuseN->Next)
						if(FuseN->Num == CN->FuseType)
							{
							CN->FuseNode = FuseN;
							break;
							}
					if(!CN->FuseNode)
						{
						ok = False;
						char errorString[200];
						unsigned char numString[10];
						strcpy(errorString, "Error: ");
						strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_DEFFILE));
						strcat(errorString, ": Can't find fusetype ");
						NumToString((SInt32)CN->FuseType, (unsigned char*)numString);
						P2CStrCat(errorString, numString);
						strcat(errorString, " for ");
						strcat(errorString, CN->Name);
						c2pstr(errorString);
						doCautionAlert((unsigned char*)errorString);

//						printf("Error: %s: Can't find fusetype %d for %s\n",
//						       (char *)GET_ARG(PI->args, ARG_DEFFILE), CN->FuseType, CN->Name);
						}
					}
				if(CN->LockType != 0)
					{
					for(LN = PI->FirstLockNode; LN; LN = LN->Next)
						if(LN->Num == CN->LockType)
							{
							CN->LockNode = LN;
							break;
							}
					if(!CN->LockNode)
						{
						ok = False;
						char errorString[200];
						unsigned char numString[10];
						strcpy(errorString, "Error: ");
						strcat(errorString, (char *)GET_avrparg(PI->avrpargs, ARG_DEFFILE));
						strcat(errorString, ": Can't find locktype ");
						NumToString((SInt32)CN->LockType, (unsigned char*)numString);
						P2CStrCat(errorString, numString);
						strcat(errorString, " for ");
						strcat(errorString, CN->Name);
						c2pstr(errorString);
						doCautionAlert((unsigned char*)errorString);
//						printf("Error: %s: Can't find locktype %d for %s\n",
//						       (char *)GET_ARG(PI->args, ARG_DEFFILE), CN->LockType, CN->Name);
						}
					}
				}
			}
		if(!ok)
			{
			FreePI(PI);
			PI = NULL;
			}
		}
	else
	{
		doCautionAlert("\pError: Unable to allocate memory!");
//		printf("Error: Unable to allocate memory\n");
	}
	return(PI);
	}


void FreePI(struct avrProgramInfo *PI)
	{
	void *Node;
	struct ChipNode *CN;
	struct VendorNode *VN;
	struct FamilyNode *FamN;
	struct ArchNode *AN;
	struct FuseNode *FuseN;
	struct FuseBitNode *FBN;
	struct LockNode *LN;
	struct LockBitNode *LBN;

	for(CN = PI->FirstChipNode; CN;)
		{
		if(CN->Name) delete CN->Name;
		Node = CN->Next;
		delete CN;
		CN = (ChipNode *) Node;
		}
	for(VN = PI->FirstVendorNode; VN;)
		{
		if(VN->Name) delete VN->Name;
		Node = VN->Next;
		delete VN;
		VN = (VendorNode*) Node;
		}
	for(FamN = PI->FirstFamilyNode; FamN;)
		{
		if(FamN->Name) delete FamN->Name;
		Node = FamN->Next;
		delete FamN;
		FamN = (FamilyNode *) Node;
		}
	for(AN = PI->FirstArchNode; AN;)
		{
		if(AN->Name) delete AN->Name;
		Node = AN->Next;
		delete AN;
		AN = (ArchNode *) Node;
		}
	for(FuseN = PI->FirstFuseNode; FuseN;)
		{
		for(FBN = FuseN->FirstFuseBitNode; FBN;)
			{
			if(FBN->ShortName) delete FBN->ShortName;
			if(FBN->LongName) delete FBN->LongName;
			Node = FBN->Next;
			delete FBN;
			FBN = (FuseBitNode *) Node;
			}
		Node = FuseN->Next;
		delete FuseN;
		FuseN = (FuseNode *) Node;
		}
	for(LN = PI->FirstLockNode; LN;)
		{
		for(LBN = LN->FirstLockBitNode; LBN;)
			{
			if(LBN->LongName) delete LBN->LongName;
			Node = LBN->Next;
			delete LBN;
			LBN = (LockBitNode *) Node;
			}
		Node = LN->Next;
		delete LN;
		LN = (LockNode *) Node;
		}
	delete PI;
	}





